﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.IO;

namespace RoboRaceUser
{
    public partial class MainForm : Form
    {
        private string server = null;
        private LogForm log = null;
        private string sessionId = null;
        private byte[,] map = null;
        private Point pos = Point.Empty;
        private char rot = 'R';
        private Bitmap robot = null;
        private string possibleMoves = null;

        public MainForm()
        {
            InitializeComponent();
            //this.ClientSize = new Size(400, 400);
            try
            {
                robot = new Bitmap("robot.png");
            }
            catch
            {
                MessageBox.Show("Nelze otevřít soubor robot.png.", "Chyba", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private string[] SendRequest(string getUrl)
        {
            try
            {
                WebRequest req = WebRequest.Create(server + getUrl);
                WebResponse res = req.GetResponse();
                StreamReader sr = new StreamReader(res.GetResponseStream());
                List<string> ret = new List<string>();
                string log = "Last Response from " +(res as HttpWebResponse).Server + "\r\n";
                while (!sr.EndOfStream)
                {
                    string row = sr.ReadLine();
                    if (!string.IsNullOrWhiteSpace(row))
                    {
                        ret.Add(row);
                        log += row + "\r\n";
                    }
                }
                sr.Close();
                res.Close();
                this.log.SetLog(log);
                return ret.ToArray();
            }
            catch (Exception error)
            {
                MessageBox.Show(error.Message, "Server Response Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
        }

        private void konecToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void novaHraToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                string[] response = SendRequest("start");
                if (response == null)
                    return;
                sessionId = response[0];
                map = new byte[20, 20];
                for (int i = 0; i < 20; ++i)
                    for (int j = 0; j < 20; ++j)
                        switch (response[i+1][j])
                        {
                            case '.':
                                map[j,i] = 0; // space
                                break;
                            case '$':
                                map[j,i] = 1; // gold
                                break;
                            case '#':
                                map[j,i] = 2; // hole
                                break;
                            case 'X':
                                map[j,i] = 3; // wall
                                break;
                            case '^':
                                map[j, i] = 5; // roller up
                                break;
                            case 'v':
                                map[j, i] = 7; // roller down
                                break;
                            case '<':
                                map[j, i] = 6; // roller left
                                break;
                            case '>':
                                map[j, i] = 4; // roller right
                                break;
                            default:
                                throw new Exception("Neznámý znak v zadaném poli!");
                        }
                string[] posstr = response[21].Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                pos = new Point(int.Parse(posstr[0]) - 1, int.Parse(posstr[1]) - 1);
                rot = posstr[2][0];
                possibleMoves = response[22];
                label2.Text = "Nabízené tahy: " + possibleMoves;
                textBox1.Enabled = true;
                button1.Enabled = true;
                pictureBox1.Invalidate();
            }
            catch (Exception error)
            {
                MessageBox.Show(error.Message, "Server Response Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            log = new LogForm();
            ServerConnect serv = new ServerConnect();
            serv.ShowDialog();
            if (serv.server == null)
                this.Close();
            else
            {
                server = serv.server;
                novaHraToolStripMenuItem_Click(sender, e);
            }
        }

        private void zmenaServeruToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Změnou serveru začnete novou hru. Opravdu chcete poračovat?", "Dotaz", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == System.Windows.Forms.DialogResult.No)
                return;
            ServerConnect serv = new ServerConnect();
            serv.ShowDialog();
            if (serv.server != null)
                server = serv.server;
            novaHraToolStripMenuItem_Click(sender, e);
        }

        private void zobrazitLogToolStripMenuItem_Click(object sender, EventArgs e)
        {
            log.Show();
        }

        private Point GetDirFromAngle(int angle)
        {
            while (angle < 0)
                angle += 360;
            while (angle >= 360)
                angle -= 360;
            switch (angle)
            {
                case 0:
                    return new Point(1, 0);
                case 90:
                    return new Point(0, -1);
                case 180:
                    return new Point(-1, 0);
                case 270:
                    return new Point(0, 1);
            }
            return Point.Empty;
        }

        private Point GoFront(int x, int y, int angle)
        {
            if ((x < 0) || (y < 0) || (x >= 20) || (y >= 20))
                return new Point(x, y);
            if ((map[x, y] == 2) || (map[x, y] == 1))
                return new Point(x, y);
            Point dir = GetDirFromAngle(angle);
            int nx = x + dir.X;
            int ny = y + dir.Y;
            if ((nx < 0) || (ny < 0) || (nx >= 20) || (ny >= 20))
                return new Point(nx, ny);
            if (map[nx, ny] == 3)
                return new Point(x, y);
            return new Point(nx, ny);
        }

        private void CheckRollers(ref int x, ref int y)
        {
            if ((x < 0) || (y < 0) || (x >= 20) || (y >= 20))
                return;
            if ((map[x, y]) > 3)
            {
                Point p = GoFront(x, y, (map[x, y] - 4) * 90);
                x = p.X;
                y = p.Y;
            }
        }

        private void pictureBox1_Paint(object sender, PaintEventArgs e)
        {
            int size = Math.Min(pictureBox1.Width, pictureBox1.Height);
            float s = size / 20.0f;
            Bitmap bmp = new Bitmap(size, size);
            Graphics g = Graphics.FromImage(bmp);
            g.Clear(Color.White);
            for (int i = 0; i < 20; ++i)
                for (int j = 0; j < 20; ++j)
                {
                    if ((pos.X == i) && (pos.Y == j))
                    {
                        g.DrawImage(robot, i * s, j * s, s - 1, s - 1);
                        switch (rot)
                        {
                            case 'N':
                                g.DrawLine(new Pen(Brushes.Blue, 3), i * s + s / 2, j * s + s / 2, i * s + s / 2, j * s); 
                                break;
                            case 'D':
                                g.DrawLine(new Pen(Brushes.Blue, 3), i * s + s / 2, j * s + s / 2, i * s + s / 2, j * s + s); 
                                break;
                            case 'R':
                                g.DrawLine(new Pen(Brushes.Blue, 3), i * s + s / 2, j * s + s / 2, i * s + s, j * s + s / 2); 
                                break;
                            case 'L':
                                g.DrawLine(new Pen(Brushes.Blue, 3), i * s + s / 2, j * s + s / 2, i * s, j * s + s / 2); 
                                break;
                        }
                    }
                    else
                        switch (map[i, j])
                        {
                            case 0:
                                break;
                            case 1:
                                g.FillEllipse(Brushes.Gold, i * s + 1, j * s + 1, s - 2, s - 2);
                                break;
                            case 2:
                                g.FillEllipse(Brushes.DarkGray, i * s + 1, j * s + 1, s - 2, s - 2);
                                break;
                            case 3:
                                g.FillRectangle(Brushes.Black, i * s, j * s, s - 1, s - 1);
                                break;
                        }
                    g.DrawRectangle(Pens.Gray, i * s, j * s, s - 1, s - 1);
                }
            if (valid)
            {
                string movs = textBox1.Text;
                int angle = 0;
                int posx = pos.X;
                int posy = pos.Y;
                int oldposx = posx;
                int oldposy = posy;
                switch (rot)
                {
                    case 'N':
                        angle = 90;
                        break;
                    case 'D':
                        angle = -90;
                        break;
                    case 'R':
                        angle = 0;
                        break;
                    case 'L':
                        angle = 180;
                        break;
                }
                Point p = Point.Empty;
                for (int i = 0; i < movs.Length; ++i)
                {
                    switch (movs[i])
                    {
                        case '1':
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                        case '2':
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                        case '3':
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                        case 'L':
                            angle += 90;
                            break;
                        case 'R':
                            angle -= 90;
                            break;
                        case 'Z':
                            p = GoFront(posx, posy, -angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                    }
                    if ((posx != oldposx) || (posy != oldposy))
                    {
                        g.DrawLine(new Pen(Brushes.DarkOrange, 3), oldposx * s + s / 2, oldposy * s + s / 2, posx * s + s / 2, posy * s + s / 2);
                        oldposx = posx;
                        oldposy = posy;
                    }
                }
            }
            g.Flush();
            e.Graphics.DrawImage(bmp, 0, 0);
        }

        private void MainForm_ResizeEnd(object sender, EventArgs e)
        {
            pictureBox1.Invalidate();
        }

        private void MainForm_ClientSizeChanged(object sender, EventArgs e)
        {
            pictureBox1.Height = this.ClientSize.Height - 80;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            textBox1_TextChanged(sender, e);
            if (!valid)
            {
                MessageBox.Show(message);
                return;
            }
            // todo: check
            string[] response = SendRequest("kolo?id=" + sessionId + "&pohyby=" + textBox1.Text);
            if (response == null)
                return;
            if (response.Length < 23)
            {
                MessageBox.Show(response[0]);
                button1.Enabled = false;
                textBox1.Enabled = false;
                label2.Text = "Hra je ukončena";
                return;
            }
            string[] posstr = response[21].Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            pos = new Point(int.Parse(posstr[0]) - 1, int.Parse(posstr[1]) - 1);
            rot = posstr[2][0];
            possibleMoves = response[22];
            label2.Text = "Nabízené tahy: " + possibleMoves;
            pictureBox1.Invalidate();
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            pictureBox1.Invalidate();
        }

        private bool valid = true;
        string message = null;
        
        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            string pmov = possibleMoves;
            string mask = "123LRZ";
            valid = true;
            if (textBox1.Text.Length > 5)
            {
                valid = false;
                message = "Obsahuje příliš mnoho znaků!";
            }
            else
            foreach (char c in textBox1.Text)
            {
                if (!mask.Contains(c))
                {
                    valid = false;
                    message = "Obsahuje neplatné znaky!";
                    break;
                }
                if (!pmov.Contains(c))
                {
                    valid = false;
                    message = "Obsahuje nepovolené příkazy!";
                    break;
                }
                pmov = pmov.Substring(0, pmov.IndexOf(c)) + ' ' + pmov.Substring(pmov.IndexOf(c) + 1);
            }
            if (valid)
            {
                label3.Text = "";
                pictureBox1.Invalidate();
            }
            else
                label3.Text = message;
        }

        private bool TestEntry(int wx, int wy, int wangle, string entry)
        {
                int angle = 0;
                int posx = pos.X;
                int posy = pos.Y;
                switch (rot)
                {
                    case 'N':
                        angle = 90;
                        break;
                    case 'D':
                        angle = -90;
                        break;
                    case 'R':
                        angle = 0;
                        break;
                    case 'L':
                        angle = 180;
                        break;
                }
                Point p = Point.Empty;
                for (int i = 0; i < entry.Length; ++i)
                {
                    switch (entry[i])
                    {
                        case '1':
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                        case '2':
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                        case '3':
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            p = GoFront(posx, posy, angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                        case 'L':
                            angle += 90;
                            break;
                        case 'R':
                            angle -= 90;
                            break;
                        case 'Z':
                            p = GoFront(posx, posy, -angle);
                            posx = p.X;
                            posy = p.Y;
                            CheckRollers(ref posx, ref posy);
                            break;
                    }
                }
                if ((angle == wangle) && (wx == posx) && (wy == posy))
                    return true;
                return false;
        }

        private string FindUsingBrutalForce(int wx, int wy, int wangle)
        {
            string amask = possibleMoves;
            for (int a = 0; a < 9; ++a)
            {
                string bmask = amask.Substring(0, a) + possibleMoves.Substring(a + 1);
                for (int b = 0; b < 8; ++b)
                {
                    string cmask = bmask.Substring(0, b) + possibleMoves.Substring(b + 1);
                    for (int c = 0; c < 7; ++c)
                    {
                        string dmask = cmask.Substring(0, c) + possibleMoves.Substring(c + 1);
                        for (int d = 0; d < 6; ++d)
                        {
                            string emask = dmask.Substring(0, d) + possibleMoves.Substring(d + 1);
                            for (int e = 0; e < 5; ++e)
                            {
                                string entry = amask[a].ToString() + bmask[b].ToString() + cmask[c].ToString() + dmask[d].ToString() + emask[e].ToString();
                                if (TestEntry(wx, wy, wangle, entry))
                                    return entry;
                            }
                        }
                    }
                }
            }
            return null;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            int wx = (int)numericUpDown1.Value - 1;
            int wy = (int)numericUpDown2.Value - 1;
            int angle = 0;
            switch (textBox2.Text.ToUpper())
            {
                case "N":
                    angle = 90;
                    break;
                case "D":
                    angle = -90;
                    break;
                case "R":
                    angle = 0;
                    break;
                case "L":
                    angle = 180;
                    break;
                default:
                    MessageBox.Show("Neplatný znak v poli směru!\n\nNápověda\nN - nahoru\nD - dolů\nR - doprava\nL - doleva", "Chybné zadání", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;;
            }
            string result = FindUsingBrutalForce(wx, wy, angle); // Brutalforce není problém - jen 126 možností
            if (result == null)
                MessageBox.Show("Nepodařilo se nalézt 5 tahů vedoucích k této pozici!");
            else
            {
                MessageBox.Show("Cesta nalezena!");
                textBox1.Text = result;
            }
        }

        private void oProgramuToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show("RoboRace - User version\nVytvořeno pro soutěž.\n\nCredits:\n  Obrázek robota ze zdroje:\n    http://nd03.jxs.cz/359/956/994714a7f4_56862754_o2.jpg");
        }
    }
}
